#include <windows.h>
#include <avisynth.h>
#include <malloc.h>
#define _USE_MATH_DEFINES
#include <math.h>

static inline unsigned char bound (int x)
{ return (unsigned)x <= 255 ? (unsigned char)x : x < 0 ? 0 : 255; }

class FDCT : public GenericVideoFilter {
protected:
	double *dct[3];
	double *init_dct (unsigned n) {
		double *dct = new double[n*n], *p = dct;
		double c1 = 1.0 / sqrt ((double)n), c2 = sqrt(2.0 / n);
		for (unsigned i=0; i<n; i++) *p++ = c1;
		for (unsigned j=1; j<n; j++)
			for (unsigned i=0,k=j; i<n; k+=2*j,i++)
				*p++ = cos (k * M_PI_2 / n) * c2;
		return dct;
	}
	FDCT (PClip c, IScriptEnvironment *e) : GenericVideoFilter(c) {
		if (!vi.IsYUY2 ())
			e->ThrowError ("DCT: input to filter must be in YUY2");
		dct[0] = init_dct (vi.height);
		dct[1] = init_dct (vi.width);
		dct[2] = init_dct (vi.width/2);
	}
public:
	~FDCT () { delete[] dct[0]; delete[] dct[1]; delete[] dct[2]; }
	static AVSValue __cdecl Create (AVSValue a, void *, IScriptEnvironment *e)
	{ return new FDCT (a[0].AsClip(), e); }
	PVideoFrame __stdcall GetFrame (int n, IScriptEnvironment *env);
};

class FDCT2 : public FDCT {
	FDCT2 (PClip c, IScriptEnvironment *e) : FDCT(c,e) { }
public:
	static AVSValue __cdecl Create (AVSValue a, void *, IScriptEnvironment *e)
	{ return new FDCT2 (a[0].AsClip(), e); }
	PVideoFrame __stdcall GetFrame (int n, IScriptEnvironment *env);
};

class IDCT : public GenericVideoFilter {
protected:
	double *dct[3];
	double *init_dct (unsigned n) {
		double *dct = new double[n*n], *p = dct;
		double c1 = 1.0 / sqrt ((double)n), c2 = sqrt(2.0 / n);
		for (unsigned i=0,j,k; i<n; i++)
			for (*p++=c1,j=1,k=0; k+=2*i+1,j<n; j++)
				*p++ = cos (k * M_PI_2 / n) * c2;
		return dct;
	}
	IDCT (PClip c, IScriptEnvironment *e) : GenericVideoFilter(c) {
		if (!vi.IsYUY2 ())
			e->ThrowError ("DCT: input to filter must be in YUY2");
		dct[0] = init_dct (vi.height);
		dct[1] = init_dct (vi.width);
		dct[2] = init_dct (vi.width/2);
	}
public:
	~IDCT () { delete[] dct[0]; delete[] dct[1]; delete[] dct[2]; }
	static AVSValue __cdecl Create (AVSValue a, void *, IScriptEnvironment *e)
	{ return new IDCT (a[0].AsClip(), e); }
	PVideoFrame __stdcall GetFrame (int n, IScriptEnvironment *env);
};

class IDCT2 : public IDCT {
	IDCT2 (PClip c, IScriptEnvironment *e) : IDCT(c,e) { }
public:
	static AVSValue __cdecl Create (AVSValue a, void *, IScriptEnvironment *e)
	{ return new IDCT2 (a[0].AsClip(), e); }
	PVideoFrame __stdcall GetFrame (int n, IScriptEnvironment *env);
};

extern "C" __declspec(dllexport) const char *__stdcall AvisynthPluginInit2 (IScriptEnvironment *env)
{
	env->AddFunction ("FDCT", "c", FDCT::Create, 0);
	env->AddFunction ("IDCT", "c", IDCT::Create, 0);
	env->AddFunction ("FDCT2", "c", FDCT2::Create, 0);
	env->AddFunction ("IDCT2", "c", IDCT2::Create, 0);
	return 0;
}

PVideoFrame __stdcall FDCT::GetFrame (int n, IScriptEnvironment *env)
{
	PVideoFrame src = child->GetFrame (n, env);
	PVideoFrame dst = env->NewVideoFrame (vi);
	unsigned char *d = dst->GetWritePtr ();
	const int dst_next = dst->GetPitch () - dst->GetRowSize ();
	const int src_next = src->GetPitch () - src->GetRowSize ();

	unsigned char *dc = d;
	double q1 = sqrt ((double)(vi.height*vi.width));
	double q2 = sqrt ((double)(vi.height*vi.width/2));
	for (unsigned i=0; (int)i<vi.height; i++) {
		double *dct1 = dct[1], *dct2 = dct[2];
		for (unsigned j=0; (int)j<vi.width; j+=2) {
			double y1 = 0, u = 0, y2 = 0, v = 0, *dct0 = &dct[0][i*vi.height];
			const unsigned char *s = src->GetReadPtr ();
			for (unsigned y=0; (int)y<vi.height; dct0++,s+=src_next,y++)
				for (unsigned x=0; (int)x<vi.width; )
					y1 += *s   * *dct0 * dct1[x],
					y2 += *s++ * *dct0 * dct1[x+vi.width],
					u  += *s++ * *dct0 * dct2[x++/2],
					y1 += *s   * *dct0 * dct1[x],
					y2 += *s++ * *dct0 * dct1[x+vi.width],
					v  += *s++ * *dct0 * dct2[x++/2];
			*d++ = (unsigned char)(y1 / q1 + 127.5);
			*d++ = (unsigned char)(u  / q2 + 127.5);
			*d++ = (unsigned char)(y2 / q1 + 127.5);
			*d++ = (unsigned char)(v  / q2 + 127.5);
			dct1 += 2 * vi.width; dct2 += vi.width/2;
		}
		d += dst_next;
	}
	dc[0] -= 127; dc[1] -= 127; dc[3] -= 127;
	return dst;
}

PVideoFrame __stdcall IDCT::GetFrame (int n, IScriptEnvironment *env)
{
	PVideoFrame src = child->GetFrame (n, env);
	PVideoFrame dst = env->NewVideoFrame (vi);
	unsigned char *d = dst->GetWritePtr ();
	const int dst_next = dst->GetPitch () - dst->GetRowSize ();
	const int src_next = src->GetPitch () - src->GetRowSize ();

	double q1 = sqrt ((double)(vi.height*vi.width));
	double q2 = sqrt ((double)(vi.height*vi.width/2));
	for (unsigned i=0; (int)i<vi.height; i++) {
		double *dct1 = dct[1], *dct2 = dct[2];
		for (unsigned j=0; (int)j<vi.width; j+=2) {
			double y1, u, y2, v, *dct0 = &dct[0][i*vi.height];
			const unsigned char *s = src->GetReadPtr ();
			for (unsigned x,y=0; x=0,(int)y<vi.height; dct0++,s+=src_next,y++) {
				if (!y) {
					double y = *s++ * *dct0;
					y1 = y * dct1[x]; y2 = y * dct1[x+vi.width];
					u  = *s++ * *dct0 * dct2[x++/2];
					y = (*s++ - 127) * *dct0;
					y1 += y * dct1[x]; y2 += y * dct1[x+vi.width];
					v  = *s++ * *dct0 * dct2[x++/2];
				}
				while ((int)x<vi.width) {
					double y = (*s++ - 127) * *dct0;
					y1 += y * dct1[x]; y2 += y * dct1[x+vi.width];
					u  += (*s++ - 127) * *dct0 * dct2[x++/2];
					y = (*s++ - 127) * *dct0;
					y1 += y * dct1[x]; y2 += y * dct1[x+vi.width];
					v  += (*s++ - 127) * *dct0 * dct2[x++/2];
				}
			}
			*d++ = bound ((int)(y1 * q1 + .5));
			*d++ = bound ((int)(u  * q2 + .5));
			*d++ = bound ((int)(y2 * q1 + .5));
			*d++ = bound ((int)(v  * q2 + .5));
			dct1 += 2 * vi.width; dct2 += vi.width / 2;
		}
		d += dst_next;
	}
	return dst;
}

static inline unsigned char quant (double c)
{ return (unsigned char)((c < 0 ? - sqrt (-c) : sqrt (c)) + 127.5); }

static inline double dequant (int c)
{ c -= 127; int x = c * c; return (double)(c < 0 ? - x : x) / 127.5; }

PVideoFrame __stdcall FDCT2::GetFrame (int n, IScriptEnvironment *env)
{
	PVideoFrame src = child->GetFrame (n, env);
	PVideoFrame dst = env->NewVideoFrame (vi);
	unsigned char *d = dst->GetWritePtr ();
	const int dst_next = dst->GetPitch () - dst->GetRowSize ();
	const int src_next = src->GetPitch () - src->GetRowSize ();

	double q1 = sqrt ((double)(vi.height*vi.width));
	double q2 = sqrt ((double)(vi.height*vi.width/2));
	double a1 = 127.5 / q1, a2 = 127.5 / q2;
	for (unsigned i=0; (int)i<vi.height; i++) {
		double *dct1 = dct[1], *dct2 = dct[2];
		for (unsigned j=0; (int)j<vi.width; j+=2) {
			double y1 = 0, u = 0, y2 = 0, v = 0, *dct0 = &dct[0][i*vi.height];
			const unsigned char *s = src->GetReadPtr ();
			for (unsigned y=0; (int)y<vi.height; dct0++,s+=src_next,y++)
				for (unsigned x=0; (int)x<vi.width; )
					y1 += *s   * *dct0 * dct1[x],
					y2 += *s++ * *dct0 * dct1[x+vi.width],
					u  += *s++ * *dct0 * dct2[x++/2],
					y1 += *s   * *dct0 * dct1[x],
					y2 += *s++ * *dct0 * dct1[x+vi.width],
					v  += *s++ * *dct0 * dct2[x++/2];
			if (i || j)
				*d++ = quant (a1 * y1), *d++ = quant (a2 * u),
				*d++ = quant (a1 * y2), *d++ = quant (a2 * v);
			else
				*d++ = (unsigned char)(y1 / q1 + .5),
				*d++ = (unsigned char)(u  / q2 + .5),
				*d++ = quant (a1 * y2),
				*d++ = (unsigned char)(v  / q2 + .5);
			dct1 += 2 * vi.width; dct2 += vi.width/2;
		}
		d += dst_next;
	}
	return dst;
}

PVideoFrame __stdcall IDCT2::GetFrame (int n, IScriptEnvironment *env)
{
	PVideoFrame src = child->GetFrame (n, env);
	PVideoFrame dst = env->NewVideoFrame (vi);
	unsigned char *d = dst->GetWritePtr ();
	const int dst_next = dst->GetPitch () - dst->GetRowSize ();
	const int src_next = src->GetPitch () - src->GetRowSize ();

	double q1 = sqrt ((double)(vi.height*vi.width));
	double q2 = sqrt ((double)(vi.height*vi.width/2));
	for (unsigned i=0; (int)i<vi.height; i++) {
		double *dct1 = dct[1], *dct2 = dct[2];
		for (unsigned j=0; (int)j<vi.width; j+=2) {
			double y1, u, y2, v, *dct0 = &dct[0][i*vi.height];
			const unsigned char *s = src->GetReadPtr ();
			for (unsigned x,y=0; x=0,(int)y<vi.height; dct0++,s+=src_next,y++) {
				if (!y) {
					double y = *s++ * *dct0;
					y1 = y * dct1[x]; y2 = y * dct1[x+vi.width];
					u  = *s++ * *dct0 * dct2[x++/2];
					y = dequant (*s++) * *dct0;
					y1 += y * dct1[x]; y2 += y * dct1[x+vi.width];
					v  = *s++ * *dct0 * dct2[x++/2];
				}
				while ((int)x<vi.width) {
					double y = dequant (*s++) * *dct0;
					y1 += y * dct1[x]; y2 += y * dct1[x+vi.width];
					u  += dequant (*s++) * *dct0 * dct2[x++/2];
					y = dequant (*s++) * *dct0;
					y1 += y * dct1[x]; y2 += y * dct1[x+vi.width];
					v  += dequant (*s++) * *dct0 * dct2[x++/2];
				}
			}
			*d++ = bound ((int)(y1 * q1 + .5));
			*d++ = bound ((int)(u  * q2 + .5));
			*d++ = bound ((int)(y2 * q1 + .5));
			*d++ = bound ((int)(v  * q2 + .5));
			dct1 += 2 * vi.width; dct2 += vi.width / 2;
		}
		d += dst_next;
	}
	return dst;
}
